def sasl_conf(args):
    sasl_mechanism = args["kafka_sasl_mechanism"].upper()

    sasl_conf = {
        "sasl.mechanism": sasl_mechanism,
        # Set to SASL_SSL to enable TLS support.
        "security.protocol": "SASL_PLAINTEXT",
    }

    if sasl_mechanism != "GSSAPI":
        sasl_conf.update(
            {
                "sasl.username": args["kafka_user_name"],
                "sasl.password": args["kafka_user_password"],
            }
        )

    if sasl_mechanism == "GSSAPI":
        sasl_conf.update(
            {
                "sasl.kerberos.service.name",
                args["kafka_user_name"],
                # Keytabs are not supported on Windows. Instead the
                # the logged on user's credentials are used to
                # authenticate.
                "sasl.kerberos.principal",
                args["kafka_user_name"],
                "sasl.kerberos.keytab",
                args["kafka_user_password"],
            }
        )

    return sasl_conf



import time

from elasticsearch import helpers
from loguru import logger


def bulk_index(es, batch: list):
    logger.info("Trying to index batch to ElasticSearch...")
    start = time.time()
    helpers.bulk(es, batch)
    end = time.time()
    logger.debug("Time taken to index: {}".format(end - start))
    logger.success("Batch successfully indexed to ElasticSearch")



import grpc
from loguru import logger
from marshmallow import EXCLUDE
from pydash import get


class CheckUserIdentity(object):
    def __init__(
        self,
        empty_response_cls,
        grpc_details="Missing or invalid user identity metadata.",
        grpc_status_code=grpc.StatusCode.UNAUTHENTICATED,
    ):
        self._empty_rsp_cls = empty_response_cls
        self._status = grpc_status_code
        self._details = grpc_details

    def __call__(self, f):
        async def wrapped_function(slf, request, context):
            meta = context.invocation_metadata()
            for item in meta:
                if item[0] == "x-user-identity" and item[1].isnumeric():
                    user = {"id": int(item[1])}
                    try:
                        return await f(slf, request, context, user)
                    except TypeError as e:
                        logger.error(e)
                        context.set_code(grpc.StatusCode.INTERNAL)
                        context.set_details(
                            f"Error: The GRPC api function {f.__name__} requires the third parameter 'user'"
                        )
                        return self._empty_rsp_cls()
                    except Exception as e:
                        raise e
            context.set_code(self._status)
            context.set_details(self._details)
            return self._empty_rsp_cls()

        return wrapped_function


# Create a class that will be used to wrap the request_dict using lodash's get function to access the field values easier
# TODO: Move this to fsai_shared_funcs
class DictGetter:
    def __init__(self, request_dict):
        self.request_dict = request_dict

    def get(self, accessor, default=None):
        return get(self.request_dict, accessor, default)

    # Create a function that plucks keys from the request_dict and rekeys them
    def pluck(self, obj_mapping):
        new_obj = {}

        for key_name, accessor in obj_mapping.items():
            # Id accessor is a list then use the first element as the accessor and the second element as the default value
            if isinstance(accessor, list):
                new_obj[key_name] = get(self.request_dict, accessor[0], accessor[1])
            else:
                new_obj[key_name] = get(self.request_dict, accessor)
        return new_obj


class ValidateRequestSchema:
    def __init__(self, schema_cls, field_name=None):
        self.schema_cls = schema_cls
        self.field_name = field_name

    def __call__(self, f):
        async def wrapped_function(slf, request_dict, context):
            if self.field_name is not None:
                # Access the field value from the request dictionary
                field_value = get(request_dict, self.field_name)

                # If field_value is None then raise an exception
                if field_value is None:
                    raise ValueError(
                        f"Field '{self.field_name}' is required in the request"
                    )

                # If field_value is not a dict then raise an exception
                if not isinstance(field_value, dict):
                    raise ValueError(
                        f"Field '{self.field_name}' must be a dict. Ensure you call ProtoToDict decorator."
                    )
            else:
                field_value = request_dict

            # Load and validate the field using the schema
            data = self.schema_cls().load(field_value, unknown=EXCLUDE)

            if self.field_name is not None:
                # Update the field value with the validated data
                request_dict[self.field_name] = data
            else:
                # Update the field value with the validated data
                request_dict = data

            # Call the original function with the modified request
            result = f(slf, DictGetter(request_dict), context)

            return await result

        return wrapped_function


class ValidateRequestSchemaGenerator:
    def __init__(self, schema_cls, field_name=None):
        self.schema_cls = schema_cls
        self.field_name = field_name

    def __call__(self, f):
        async def wrapped_function(slf, request_dict, context):
            if self.field_name is not None:
                # Access the field value from the request dictionary
                field_value = get(request_dict, self.field_name)

                # If field_value is None then raise an exception
                if field_value is None:
                    raise ValueError(
                        f"Field '{self.field_name}' is required in the request"
                    )

                # If field_value is not a dict then raise an exception
                if not isinstance(field_value, dict):
                    raise ValueError(
                        f"Field '{self.field_name}' must be a dict. Ensure you call ProtoToDict decorator."
                    )
            else:
                field_value = request_dict

            # Load and validate the field using the schema
            data = self.schema_cls().load(field_value, unknown=EXCLUDE)

            if self.field_name is not None:
                # Update the field value with the validated data
                request_dict[self.field_name] = data
            else:
                # Update the field value with the validated data
                request_dict = data

            # Call the original function with the modified request
            result = f(slf, DictGetter(request_dict), context)

            if hasattr(
                result, "__aiter__"
            ):  # Check if result is an async iterator (yields)
                async for item in result:
                    yield item
            else:  # Result is not an async iterator, assumes it returns a value
                raise Exception(
                    "ValidateRequestSchemaGenerator should only be used on functions that return an async iterator"
                )

        return wrapped_function



import pickle

import redis


# Write a class that extends redis.Redis and add the methods you need to override set and get to serialize and deserialize using pickle
class FSAIRedis(redis.Redis):
    def pset(self, name, value, *args, **kwargs):
        return super().set(name, pickle.dumps(value), *args, **kwargs)

    def pget(self, name, *args, **kwargs):
        value = super().get(name, *args, **kwargs)
        if value is None:
            return value
        return pickle.loads(value)



from datetime import datetime

from fsai_shared_funcs.proto_helpers.json_format import MessageToDict, ParseDict


def dict_to_proto(input_dict, proto_class, ignore_unknown_fields=True):   
    # Perform conversion of values to match ParseDict's requirements
    for key, value in input_dict.items():
        # Convert to .isoformat("T") + "Z" to match the format of the protobuf timestamp
        if isinstance(value, datetime):
            input_dict[key] = value.isoformat("T") + "Z"

    return ParseDict(input_dict, proto_class(), ignore_unknown_fields=ignore_unknown_fields)


def dict_to_proto_list(input_list, proto_class, ignore_unknown_fields=True):
    """
    Convert a list of dictionaries to a list of protobuf messages.
    
    Args:
        input_list (list[dict]): List of dictionaries to be converted.
        proto_class (protobuf message class): The protobuf message class to convert to.
        ignore_unknown_fields (bool, optional): Whether to ignore unknown fields when parsing. Default is True.
    
    Returns:
        list[protobuf message]: List of protobuf messages.
    """
    return [dict_to_proto(item, proto_class, ignore_unknown_fields=ignore_unknown_fields) for item in input_list]

    
def proto_to_dict(input_proto, including_default_value_fields=True, preserving_proto_field_name=True, float_precision=14):
    """
    Convert a protobuf message to a dictionary.
    
    Args:
        input_proto (protobuf message): The protobuf message to be converted.
        including_default_value_fields (bool, optional): Whether to include fields that have default values. Default is True.
        preserving_proto_field_name (bool, optional): Whether to preserve the original proto field names. Default is True.
        float_precision (int, optional): The number of decimal places to preserve for floats. Default is 14.
    
    Returns:
        dict: The protobuf message as a dictionary.
    """
    output_dict = MessageToDict(
        input_proto,
        including_default_value_fields=including_default_value_fields,
        preserving_proto_field_name=preserving_proto_field_name,
        float_precision=float_precision,
    )

    # Perform conversion of values to match ParseDict's requirements
    for key, value in output_dict.items():
        # Convert to datetime object
        if isinstance(value, str) and len(value) == 27 and value[-1] == "Z":
            try:
                output_dict[key] = datetime.strptime(value, '%Y-%m-%dT%H:%M:%S.%fZ')
            except Exception as e:
                pass

    return output_dict
